(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Arith.
Require Import MinMax.

Require Import predicate.
Require Import carrier.
Require Import maps.

Definition finite {X} (P : X -> Prop) : Prop := exists l, carrier l ~ P.
Notation " 'finite_type' X " := (finite (fun _ : X => True)) (at level 50).

Section Finite.

  Variable X : Type.
  Hypothesis eqX_dec : forall x y : X, { x = y } + { x <> y }.

  Proposition finite_id (A B : X -> Prop) : A ~ B -> finite A -> finite B.
  Proof. intros ? [ l Hl ]; exists l; apply id_transitive with (1 := Hl); auto. Qed.

  Proposition finite_carrier (l : list X) : finite (carrier l).
  Proof. exists l; auto. Qed.

  Proposition finite_empty : finite (X:=X) empty.
  Proof. exists nil; auto. Qed.

  Proposition finite_sg (a : X) : finite (sg a).
  Proof. 
    exists (a::nil).
    apply id_transitive with (1 := carrier_cons _ _ _).
    split; pred_solve.
  Qed.

  Proposition finite_union (A B : X -> Prop) : finite A -> finite B -> finite (A cup B).
  Proof.
    intros [ la ] [ lb ]; exists (la++lb).
    apply id_transitive with (1 := carrier_app _ _ _); auto.
  Qed.

  Definition bool_dec (P : X -> Prop) : Prop := exists f : X -> bool, P ~ (fun x => f x = true).

  Proposition bool_dec_imp_dec P : bool_dec P -> forall x, P x \/ ~ P x.
  Proof.
    intros [ f [ Hf1 Hf2 ] ] x.
    case_eq (f x); intros H.
    apply Hf2 in H; auto.
    right; intros H'; apply Hf1 in H'.
    rewrite -> H in H'; discriminate H'.
  Qed.

  Proposition bool_dec_not P : bool_dec P -> bool_dec (fun x => ~ P x).
  Proof.
    intros [ f [ Hf1 Hf2 ] ].
    exists (fun x => negb (f x) ).
    unfold negb.
    split; intros x Hx.
    case_eq (f x); auto.
    intros H; apply Hf1 in H.
    destruct (f x).
    discriminate Hx.
    discriminate H.
  Qed.

  Proposition finite_bool_dec P : finite P -> bool_dec P. 
  Proof.
    intros [ l [ Hl1 Hl2 ] ].
    exists (fun x => match (carrier_dec _ eqX_dec l x) with left _ => true | _ => false end).
    split; intros x Hx. 
    apply Hl2 in Hx. destruct (carrier_dec _ eqX_dec l x) as [ _ | H ]; auto; contradiction (H Hx).
    apply Hl1; destruct (carrier_dec _ eqX_dec l x); auto; discriminate Hx.
  Qed.

  Hint Resolve bool_dec_imp_dec finite_bool_dec.

  Proposition finite_in_dec (A : X -> Prop) : finite A -> forall x, A x \/ ~ A x.
  Proof. intros; auto. Qed.
 
  Proposition finite_inter_bool_dec A B : finite A -> bool_dec B -> finite (A cap B).
  Proof.
    intros [ la HA ] [ fb HB ].
    exists (filter fb la).
    apply id_transitive with (1 := carrier_filter _ _ _).
    auto.
  Qed.

  Hint Resolve finite_inter_bool_dec.

  Proposition finite_inter (A B : X -> Prop) : finite A -> finite B -> finite (A cap B).
  Proof. intros ? HB; apply finite_bool_dec in HB; auto. Qed.

  Hint Resolve finite_inter_bool_dec finite_inter.

  Definition comp_sg (x : X) y := x <> y.
  
  Proposition comp_sg_bool_dec x : bool_dec (comp_sg x).
  Proof.
    apply bool_dec_not, finite_bool_dec.
    exists (x::nil); split; auto.
    intros y Hy; rewrite Hy; auto.
  Qed.

  Hint Immediate comp_sg_bool_dec.
  
  Definition without_sg A x := A cap (comp_sg x).

  Proposition without_sg_finite A x : finite A -> finite (without_sg A x).
  Proof. unfold without_sg; intros; auto. Qed.
 
  Proposition without_sg_in (A : _ -> Prop) x y : A y -> x <> y -> without_sg A x y.
  Proof. intros; apply bin_inter_sg'; auto. Qed.

  Hint Resolve without_sg_finite without_sg_in.

  Proposition without_sg_out (A : _ -> Prop) x y : without_sg A x y -> A y /\ x <> y.
  Proof. intros H; apply bin_inter_sg in H; auto. Qed.

  Proposition without_sg_cup A x : A inc without_sg A x cup sg x.
  Proof. intros y ?; destruct (eqX_dec x y); pred_solve. Qed.

  Proposition without_sg_inc A x : without_sg A x inc A.
  Proof. intros ? Hy; apply without_sg_out in Hy; tauto. Qed.

  Hint Immediate without_sg_cup without_sg_inc.

  Proposition without_sg_inc_l A B x : A inc B cup (sg x) -> without_sg A x inc B.
  Proof.
    intros H y Hy.
    apply without_sg_out in Hy.
    destruct Hy as [ Hy1 Hy2 ].
    apply H in Hy1.
    pred_solve.
    contradiction (Hy2 Hy1).
  Qed.
   
  Definition without_pair A x y := without_sg (without_sg A x) y.
  
  Proposition without_pair_finite A x y : finite A -> finite (without_pair A x y).
  Proof. unfold without_pair; auto. Qed.

  Proposition without_pair_in (A : _ -> Prop) x y z : A z -> x <> z -> y <> z -> without_pair A x y z.
  Proof. unfold without_pair; auto. Qed. 

  Proposition without_pair_out (A : _ -> Prop) x y z : without_pair A x y z -> A z /\ x <> z /\ y <> z.
  Proof. 
    unfold without_pair; intro H.
    apply without_sg_out in H.
    destruct H as [ H ? ].
    apply without_sg_out in H.
    tauto.
  Qed.

  Hint Resolve without_pair_finite without_pair_in.

  Proposition without_pair_cup A x y : A inc without_pair A x y cup (sg x cup sg y).
  Proof.
    intros z ?.
    destruct (eqX_dec x z); pred_solve.
    destruct (eqX_dec y z); pred_solve.
  Qed.

  Proposition without_pair_inc A x y : without_pair A x y inc A.
  Proof. intros ? Hz; apply without_pair_out in Hz; tauto. Qed.

  Hint Immediate without_pair_cup without_pair_inc.

  Proposition without_pair_inc_l A B x y : A inc B cup (sg x cup sg y) -> without_pair A x y inc B.
  Proof.
    intros H z Hz.
    apply without_pair_out in Hz.
    destruct Hz as ( Hz1 & Hz2 & Hz3 ).
    apply H in Hz1.
    pred_solve.
    contradiction (Hz2 Hz1).
    contradiction (Hz3 Hz1).
  Qed.

  Proposition finite_decomp (A B C : X -> Prop) : finite A 
                                               -> finite B 
                                               -> A inc (C cup B) 
                                               -> { C' : _ 
                                                & { B' |  finite C' 
                                                       /\ finite B'  
                                                       /\ C' inc C 
                                                       /\ B' inc B 
                                                       /\ A ~ C' cup B' } }.
  Proof.
    intros HA HB H.
    generalize (finite_bool_dec _ HA) (finite_bool_dec _ HB) (finite_bool_dec _ HB).
    intros bdA bdB bdnB. apply bool_dec_not in bdnB.
    exists (A  cap (fun x => ~ B x)). 
    exists (A cap B).
    split; auto.
    split; auto.
    split.

    intros x Hx; pred_solve.
    apply H in H0.
    pred_solve; tauto.

    split; auto.

    apply id_transitive with (2 := bin_inter_distrib_right _ _ _ _).
    split; auto.
    intros x ?;
    destruct (bool_dec_imp_dec _ bdB x);
    pred_solve.
  Qed.

  Let lArec (lB : list X) : forall A, (forall x, A x \/ ~ A x) -> A inc carrier lB -> finite A.
  Proof.
    induction lB as [ | b lB ].
    exists nil.
    split; auto.
 
    intros A Adec HA.
    set (A' x := A x /\ x <> b).
    assert (forall x, A' x \/ ~ A' x) as A'dec.
      unfold A'; intros x;
      destruct (eqX_dec x b); destruct (Adec x); tauto.

    destruct (IHlB _ A'dec) as ( lA' & H' ); auto.
    
    intros x [ Hx1 Hx2 ].
    apply HA in Hx1.
    simpl in Hx1.
    destruct Hx1; auto.
    contradict Hx2; auto.
 
    destruct (Adec b) as [ Hb | Hb ].

    exists (b::lA').
    split.
    intros x [ Hx | Hx ].
    rewrite <- Hx; auto.
    apply H' in Hx.
    destruct Hx; auto.
    intros x Hx.
    destruct (eqX_dec x b) as [ H | H ].
    rewrite H; auto.
    right; apply H'; split; auto.

    exists (lA').
    apply id_transitive with (1 := H').
    unfold A'; split; intros x Hx; try tauto.
    split; auto.
    intros H; rewrite H in Hx; contradict Hb; auto.
  Qed.

  Proposition finite_dec_inc (A B : X -> Prop) : (forall x, A x \/ ~ A x) -> A inc B -> finite B -> finite A.
  Proof.
    intros H1 H2 ( lB & HlB1 & HlB2).
    apply lArec with (lB := lB); auto.
    apply inc_transitive with (1 := H2); auto.
  Qed.
      
End Finite.

Section Finite_map.

  Variable X Y : Type.
  Variable f : X -> Y.

  Variable P : X -> Prop.
  Hypothesis Pfin : finite P.

  Proposition finite_pd_image : finite (pd_image f P).
  Proof.
    destruct Pfin as [ l Hk ].
    exists (map f l).
    destruct Hk as [ Hk Hk' ].
    split; intros y Hy.
    destruct (proj1 (in_map_iff f l y) Hy) as [ x Hx ].
    destruct Hx as [ Hx Hx' ].
    exists x. apply Hk in Hx'. tauto.
    destruct Hy as [ x Hx ].
    destruct Hx as [ Hx Hx' ].
    rewrite <- Hx'. apply in_map.
    apply Hk'. trivial.
  Qed.
    
End Finite_map.

Section finite_inc_seq.

  (* if Q is a finite subset of the limit of an increasing sequence of predicates
     then Q is a subset of some element of the sequence
  *)

  Variable X : Type.
  Variable P : nat -> X -> Prop.
  Hypothesis Pinc : forall n, P n inc P (S n).

  Let Pinc' n m : P n inc P (m + n).
  Proof.
    induction m.
    auto.
    apply inc_transitive with (1 := IHm).
    apply Pinc.
  Qed.
  
  Theorem seq_inc n m : n <= m -> P n inc P m.
  Proof.
    intros H.
    cutrewrite (m = (m-n)+n).
    auto.
    rewrite plus_comm.
    apply le_plus_minus; auto.
  Qed.

  Let finite_union_rec l : carrier l inc union P -> exists n, carrier l inc P n.
  Proof.
    induction l as [ | l Hl ]; intros H; auto.
    exists 0; auto.
    apply carrier_cons_1 in H.
    destruct H as [ H1 H2 ].
    apply union_sg in H1.
    destruct H1 as [ x Hx ].
    apply IHHl in H2.
    destruct H2 as [ m H2 ].
    exists (max x m).
    apply carrier_cons_2.
    apply seq_inc with (n := x); auto.
    apply le_max_l.
    apply inc_transitive with (1 := H2).
    apply seq_inc, le_max_r.
  Qed.

  Theorem finite_union_seq Q : finite Q -> Q inc union P -> exists n, Q inc P n.
  Proof.
    intros [ l H1 ] H2.
    destruct (finite_union_rec l) as [ n Hn ].
    apply inc_transitive with (2 := H2), H1.
    exists n.
    apply inc_transitive with (2 := Hn), H1.
  Qed.

End finite_inc_seq.

Section Finite_nat.

  Variable P : nat -> Prop.

  Let max_list := 
    let fix loop l { struct l } := match l with
        | nil => O
        | x::l => max x (loop l)
    end in loop.

  Let max_prop l : forall x, (carrier l) x -> x <= max_list l.
  Proof.
    induction l; intros x Hx.
    simpl in Hx. contradiction Hx.
    simpl in Hx |- *. destruct Hx as [ Hx | Hx ].
    rewrite <- Hx. apply le_max_l.
    apply IHl in Hx. apply (le_trans _ (max_list l)).
    trivial. apply le_max_r.
  Qed.

  Proposition finite_nat_bounded : finite P -> exists k, forall x, P x -> x <= k.
  Proof.
    intros H.
    destruct H as [ l Hl ].
    exists (max_list l).
    intros x Hx.
    apply (proj2 Hl) in Hx.
    apply max_prop.
    trivial.
  Qed.

End Finite_nat.

Implicit Arguments without_sg [ X ].
Implicit Arguments without_pair [ X ].

Hint Immediate finite_carrier finite_empty finite_sg finite_pd_image.
Hint Resolve finite_id finite_union finite_inter.

Hint Resolve without_sg_finite without_sg_in.
Hint Immediate without_sg_cup without_sg_inc.
Hint Resolve without_pair_finite without_pair_in.
Hint Immediate without_pair_cup without_pair_inc.



  
